==> lpc4.05.09.tar.gz: lpc4/src/mapping.c <==
#include <string.h>
#ifdef sun
#include <alloca.h>
#endif
#include "config.h"
#include "lint.h"
#include "array.h"
#include "interpret.h"
#include "object.h"
#include "wiz_list.h"
#include "regexp.h"
#include "exec.h"
#include "main.h"

extern struct vector *insert_alist PROT((struct svalue *key,
					 struct svalue *key_data,
					 struct vector *list));
extern int alist_cmp PROT((struct svalue *,struct svalue *));
extern int assoc PROT((struct svalue *, struct vector *));

extern struct svalue const0,const_undefined;

struct vector *allocate_mapping(struct vector *v,struct vector *w)
{
  struct vector *m;

  m = allocate_array(3);
  m->item[0].u.vec = v;
  m->item[0].type = T_ALIST_PART;
  m->item[1].u.vec = w;
  m->item[1].type = T_ALIST_PART;
  m->item[2]=const_undefined;	/* Structure assign... */
  return m;
}

INLINE void map_set_default(struct vector *v,struct svalue *s)
{ assign_svalue(&(v->item[2]),s); }

INLINE struct svalue *map_get_default(struct vector *v)
{ return &(v->item[2]); }

INLINE void free_mapping(struct vector *m)
{ free_vector(m); }

struct svalue *get_map_lvalue(struct vector *m,
			      struct svalue *index,
			      int needlval)
{
  int i;

  i = assoc(index, m->item[0].u.vec);
  if (i < 0)
  {
    if (needlval)
    {
      /*
       * NOTE: we change the mapping here!
       */
      m = insert_alist(index, &const0, m);
      i = assoc(index, m->item[0].u.vec);
    }else{
      return &(m->item[2]);
    }
  }
  if(IS_ZERO(m->item[1].u.vec->item+i) &&
     m->item[1].u.vec->item[i].subtype==NUMBER_UNDEFINED)
    m->item[1].u.vec->item[i].subtype=NUMBER_NUMBER;

  return & m->item[1].u.vec->item[i];
}

struct vector *remove_mapping2(struct vector *m, struct svalue *s)
{
  int i;

  i = assoc(s, m->item[0].u.vec);
  m->ref++;
  if(i>=0) donk_alist_item(m,i);
  return m;
}

struct vector *filter_mapping (struct vector *p,
			       char *func,
			       struct object *ob,
			       struct svalue *extra)
{
    struct vector *q, *r;
    struct svalue *v;
    char *flags;
    int cnt,res;
    
    res=0;
    r=0;
    if ( !func || !ob || (ob->flags & O_DESTRUCTED)) {
	return 0;
    }
    if (p->item[0].u.vec->size<1) {
	p = allocate_array(0);
	return allocate_mapping(p, p);
    }

    q = p->item[1].u.vec;
    flags=xalloc(q->size+1); 
    for (cnt=0;cnt<q->size;cnt++) {
	flags[cnt]=0;
	push_svalue(&q->item[cnt]);
	if (extra) {
	    push_svalue(extra);
	    v = apply (func, ob, 2);
	} else {
	    v = apply (func, ob, 1);
	}
	if ((v) && (v->type==T_NUMBER)) {
	    if (v->u.number) { flags[cnt]=1; res++; }
	}
    }
    q = allocate_array(res);
    r = allocate_array(res);
    if (res) {
	for (cnt = res = 0; res < q->size; cnt++) {
	    if (flags[cnt]) {
		assign_svalue_no_free (&q->item[res],
				       &p->item[0].u.vec->item[cnt]);
		assign_svalue_no_free (&r->item[res++],
				       &p->item[1].u.vec->item[cnt]);
	    }
	}
    }
    free(flags);
    return allocate_mapping(q, r);
}

struct vector *map_mapping (struct vector *arr,
			    char *func,
			    struct object *ob,
			    struct svalue *extra)
{
    struct vector *q, *r;
    struct svalue *v;
    int cnt;
    
    if (arr->size < 1) {
	arr = allocate_array(0);
	return allocate_mapping(arr, arr);
    }

    q = arr->item[0].u.vec;
    q = slice_array(q, 0, q->size - 1); /* make a copy */
    arr = arr->item[1].u.vec;
    r = allocate_array(arr->size);
    
    for (cnt = 0; cnt < arr->size; cnt++)
    {
	if (ob->flags & O_DESTRUCTED)
	    error("object used by map_mapping destructed",0,0,0,0,0,0,0,0,0); /* amylaar */
	push_svalue(&q->item[cnt]);  /* Modified to use both index and value */
	push_svalue(&arr->item[cnt]);

	if (extra)
	{
	    push_svalue (extra);
	    v = apply(func, ob, 3);
	}
	else 
	{
	    v = apply(func,ob,2);
	}
	if (v)
	    assign_svalue_no_free (&r->item[cnt], v);
    }
    return allocate_mapping(q, r);
}

/* Is it a bird?
 * A Boing 747?
 * A flying whale?
 * no, it's Profezzorn's latest hack....
 */

struct vector *intersect_mapping(struct vector *a,struct vector *b)
{
  int q,w,d,num;
  struct vector *ret;

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(d<0) q++;
    else if(d>0) w++;
    else num++,q++,w++;
  }
  ret=allocate_mapping(allocate_array(num),allocate_array(num));

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(d<0) q++;
    else if(d>0) w++;
    else { mutilate_mapping(ret,num++,a,q++); b++; }
  }

  return ret;
}

struct vector *subtract_mapping(struct vector *a,struct vector *b)
{
  int q,w,d,num;
  struct vector *ret;

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(!d) q++;
    else if(d>0) w++;
    else num++,q++;
  }
  num+=a->item[0].u.vec->size-q;
  ret=allocate_mapping(allocate_array(num),allocate_array(num));

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(!d) q++;
    else if(d>0) w++;
    else mutilate_mapping(ret,num++,a,q++);
  }

  while(q<a->item[0].u.vec->size) mutilate_mapping(ret,num++,a,q++);
  return ret;  
}

/* This one just takes two mappings, sorry */

#define SUM_MAP_ERR(X) \
        { \
          free_vector(indices); \
          free_vector(values); \
          error((X),0,0,0,0,0,0,0,0,0); \
        }

struct vector *sum_mappings(
	struct vector *mapa,
	struct vector *mapb,
	struct svalue *funa,
	struct svalue *funab,
	struct svalue *funb)
{
  int size,ap,bp,asize,bsize,tmp;
  struct svalue *aa,*bb,*ip,*vp,*ret;
  struct vector *indices;
  struct vector *values;

  asize=mapa->item[0].u.vec->size;
  bsize=mapb->item[0].u.vec->size;
  aa=& mapa->item[0].u.vec->item[0];
  bb=& mapb->item[0].u.vec->item[0];
  ap=bp=size=0;

  while(ap<bsize && bp<bsize)
  {
    size++;
    tmp=alist_cmp(aa+ap,bb+bp);

    if(tmp<0)		{ if(funa->type==T_FUNCTION) ap++; }
    else if(tmp>0)	{ if(funb->type==T_FUNCTION) bp++; }
    else		{ if(funab->type==T_FUNCTION) ap++,bp++; }
  }
  size+=asize-ap+bsize-bp;
  indices=allocate_array(size);
  values=allocate_array(size);
  ip=& indices->item[0];
  vp=& values->item[0];
  ap=bp=size=0;

  while(ap<bsize && bp<bsize)
  {
    tmp=alist_cmp(& aa[ap],& bb[bp]);
    if(!tmp)
    {
      if(funab->type==T_FUNCTION)
      {
        push_svalue(&mapa->item[1].u.vec->item[ap++]);
        push_svalue(&mapb->item[1].u.vec->item[bp++]);
        if(funab->u.ob->flags & O_DESTRUCTED)
          SUM_MAP_ERR("Object used by sum_mappings destructed.\n");

        ret=apply_lambda(funab,2);
        if(!ret) SUM_MAP_ERR("Function used by sum_mappings() not found.\n");

        assign_svalue(&ip[size],&aa[ap]);
        assign_svalue(&vp[size++],ret);
      }
    }else if(tmp<0){
      if(funa->type==T_FUNCTION)
      {
        push_svalue(&mapa->item[1].u.vec->item[ap++]);
        if(funa->u.ob->flags & O_DESTRUCTED)
          SUM_MAP_ERR("Object used by sum_mappings destructed.\n");

        ret=apply_lambda(funa,1);
        if(!ret) SUM_MAP_ERR("Function used by sum_mappings() not found.\n");

        assign_svalue(&ip[size],&aa[ap]);
        assign_svalue(&vp[size++],ret);
      }
    }else{
      if(funb->type==T_FUNCTION)
      {
        push_svalue(&mapb->item[1].u.vec->item[bp++]);
          SUM_MAP_ERR("Object used by sum_mappings destructed.\n");

        ret=apply_lambda(funb,1);
        if(!ret) SUM_MAP_ERR("Function used by sum_mappings() not found.\n");

        assign_svalue(&ip[size],&bb[bp]);
        assign_svalue(&vp[size++],ret);
      }
    }
  }
  return allocate_mapping(indices,values);
}

==> lpc4.05.10.tar.gz: lpc4/src/mapping.c <==
#include "global.h"
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#include "array.h"
#include "interpret.h"
#include "object.h"
#include "wiz_list.h"
#include "regexp.h"
#include "exec.h"
#include "main.h"
#include "list.h"
#include "simulate.h"

extern struct svalue const0,const_undefined;

struct vector *allocate_mapping(struct vector *v,struct vector *w)
{
  struct vector *m;

  m = allocate_array(3);
  m->item[0].u.vec = v;
  m->item[0].type = T_ALIST_PART;
  m->item[1].u.vec = w;
  m->item[1].type = T_ALIST_PART;
  m->item[2]=const_undefined;	/* Structure assign... */
  return m;
}

INLINE void map_set_default(struct vector *v,struct svalue *s)
{ assign_svalue(&(v->item[2]),s); }

INLINE struct svalue *map_get_default(struct vector *v)
{ return &(v->item[2]); }

INLINE void free_mapping(struct vector *m)
{ free_vector(m); }

struct svalue *get_map_lvalue(struct vector *m,
			      struct svalue *index,
			      int needlval)
{
  int i;

  i = assoc(index, m->item[0].u.vec);
  if (i < 0)
  {
    if (needlval)
    {
      /*
       * NOTE: we change the mapping here!
       */
      m = insert_alist(index, &const0, m);
      i = assoc(index, m->item[0].u.vec);
    }else{
      return &(m->item[2]);
    }
  }
  if(IS_ZERO(m->item[1].u.vec->item+i) &&
     m->item[1].u.vec->item[i].subtype==NUMBER_UNDEFINED)
    m->item[1].u.vec->item[i].subtype=NUMBER_NUMBER;

  return & m->item[1].u.vec->item[i];
}

struct vector *remove_mapping2(struct vector *m, struct svalue *s)
{
  int i;

  i = assoc(s, m->item[0].u.vec);
  m->ref++;
  if(i>=0) donk_alist_item(m,i);
  return m;
}

struct vector *filter_mapping (struct vector *p,
			       char *func,
			       struct object *ob,
			       struct svalue *extra)
{
    struct vector *q, *r;
    struct svalue *v;
    char *flags;
    int cnt,res;
    
    res=0;
    r=0;
    if ( !func || !ob || (ob->flags & O_DESTRUCTED)) {
	return 0;
    }
    if (p->item[0].u.vec->size<1) {
	p = allocate_array(0);
	return allocate_mapping(p, p);
    }

    q = p->item[1].u.vec;
    flags=xalloc(q->size+1); 
    for (cnt=0;cnt<q->size;cnt++) {
	flags[cnt]=0;
	push_svalue(&q->item[cnt]);
	if (extra) {
	    push_svalue(extra);
	    v = apply (func, ob, 2);
	} else {
	    v = apply (func, ob, 1);
	}
	if ((v) && (v->type==T_NUMBER)) {
	    if (v->u.number) { flags[cnt]=1; res++; }
	}
    }
    q = allocate_array(res);
    r = allocate_array(res);
    if (res) {
	for (cnt = res = 0; res < q->size; cnt++) {
	    if (flags[cnt]) {
		assign_svalue_no_free (&q->item[res],
				       &p->item[0].u.vec->item[cnt]);
		assign_svalue_no_free (&r->item[res++],
				       &p->item[1].u.vec->item[cnt]);
	    }
	}
    }
    free(flags);
    return allocate_mapping(q, r);
}

struct vector *map_mapping (struct vector *arr,
			    char *func,
			    struct object *ob,
			    struct svalue *extra)
{
    struct vector *q, *r;
    struct svalue *v;
    int cnt;
    
    if (arr->size < 1) {
	arr = allocate_array(0);
	return allocate_mapping(arr, arr);
    }

    q = arr->item[0].u.vec;
    q = slice_array(q, 0, q->size - 1); /* make a copy */
    arr = arr->item[1].u.vec;
    r = allocate_array(arr->size);
    
    for (cnt = 0; cnt < arr->size; cnt++)
    {
	if (ob->flags & O_DESTRUCTED)
	    error("object used by map_mapping destructed"); /* amylaar */
	push_svalue(&q->item[cnt]);  /* Modified to use both index and value */
	push_svalue(&arr->item[cnt]);

	if (extra)
	{
	    push_svalue (extra);
	    v = apply(func, ob, 3);
	}
	else 
	{
	    v = apply(func,ob,2);
	}
	if (v)
	    assign_svalue_no_free (&r->item[cnt], v);
    }
    return allocate_mapping(q, r);
}

/* Is it a bird?
 * A Boing 747?
 * A flying whale?
 * no, it's Profezzorn's latest hack....
 */

struct vector *intersect_mapping(struct vector *a,struct vector *b)
{
  int q,w,d,num;
  struct vector *ret;

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(d<0) q++;
    else if(d>0) w++;
    else num++,q++,w++;
  }
  ret=allocate_mapping(allocate_array(num),allocate_array(num));

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(d<0) q++;
    else if(d>0) w++;
    else { mutilate_mapping(ret,num++,a,q++); b++; }
  }

  return ret;
}

struct vector *subtract_mapping(struct vector *a,struct vector *b)
{
  int q,w,d,num;
  struct vector *ret;

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(!d) q++;
    else if(d>0) w++;
    else num++,q++;
  }
  num+=a->item[0].u.vec->size-q;
  ret=allocate_mapping(allocate_array(num),allocate_array(num));

  num=q=w=0;

  while(q<a->item[0].u.vec->size && w<b->item[0].u.vec->size)
  {
    d=alist_cmp(&(a->item[0].u.vec->item[q]),&(b->item[0].u.vec->item[w]));
    if(!d) q++;
    else if(d>0) w++;
    else mutilate_mapping(ret,num++,a,q++);
  }

  while(q<a->item[0].u.vec->size) mutilate_mapping(ret,num++,a,q++);
  return ret;  
}

/* This one just takes two mappings, sorry */

#define SUM_MAP_ERR(X) \
        { \
          free_vector(indices); \
          free_vector(values); \
          error((X)); \
        }

struct vector *sum_mappings(
	struct vector *mapa,
	struct vector *mapb,
	struct svalue *funa,
	struct svalue *funab,
	struct svalue *funb)
{
  int size,ap,bp,asize,bsize,tmp;
  struct svalue *aa,*bb,*ip,*vp,*ret;
  struct vector *indices;
  struct vector *values;

  asize=mapa->item[0].u.vec->size;
  bsize=mapb->item[0].u.vec->size;
  aa=& mapa->item[0].u.vec->item[0];
  bb=& mapb->item[0].u.vec->item[0];
  ap=bp=size=0;

  while(ap<bsize && bp<bsize)
  {
    size++;
    tmp=alist_cmp(aa+ap,bb+bp);

    if(tmp<0)		{ if(funa->type==T_FUNCTION) ap++; }
    else if(tmp>0)	{ if(funb->type==T_FUNCTION) bp++; }
    else		{ if(funab->type==T_FUNCTION) ap++,bp++; }
  }
  size+=asize-ap+bsize-bp;
  indices=allocate_array(size);
  values=allocate_array(size);
  ip=& indices->item[0];
  vp=& values->item[0];
  ap=bp=size=0;

  while(ap<bsize && bp<bsize)
  {
    tmp=alist_cmp(& aa[ap],& bb[bp]);
    if(!tmp)
    {
      if(funab->type==T_FUNCTION)
      {
        push_svalue(&mapa->item[1].u.vec->item[ap++]);
        push_svalue(&mapb->item[1].u.vec->item[bp++]);
        if(funab->u.ob->flags & O_DESTRUCTED)
          SUM_MAP_ERR("Object used by sum_mappings destructed.\n");

        ret=apply_lambda(funab,2);
        if(!ret) SUM_MAP_ERR("Function used by sum_mappings() not found.\n");

        assign_svalue(&ip[size],&aa[ap]);
        assign_svalue(&vp[size++],ret);
      }
    }else if(tmp<0){
      if(funa->type==T_FUNCTION)
      {
        push_svalue(&mapa->item[1].u.vec->item[ap++]);
        if(funa->u.ob->flags & O_DESTRUCTED)
          SUM_MAP_ERR("Object used by sum_mappings destructed.\n");

        ret=apply_lambda(funa,1);
        if(!ret) SUM_MAP_ERR("Function used by sum_mappings() not found.\n");

        assign_svalue(&ip[size],&aa[ap]);
        assign_svalue(&vp[size++],ret);
      }
    }else{
      if(funb->type==T_FUNCTION)
      {
        push_svalue(&mapb->item[1].u.vec->item[bp++]);
          SUM_MAP_ERR("Object used by sum_mappings destructed.\n");

        ret=apply_lambda(funb,1);
        if(!ret) SUM_MAP_ERR("Function used by sum_mappings() not found.\n");

        assign_svalue(&ip[size],&bb[bp]);
        assign_svalue(&vp[size++],ret);
      }
    }
  }
  return allocate_mapping(indices,values);
}
